#include <iostream.h>

#include "car.h"
#include "user.h"
#include "sim.h"
#include "track.h"

#include <GL/gl.h>
#include <GL/glu.h>
#include "glaux.h"

#include "lib/lib_math.h"

wheel wheels[4];

/////////////////////////////////////////////
// returns tire patch normalized vector w.r.t. ground
/////////////////////////////////////////////
bool
wheel::get_patch_vel( const VEC& car_vel, VEC car_forward, const VEC& car_up ) {

  VEC car_right;
  car_right.xprod( car_forward, car_up );
  car_forward.smult( cos( DEGS_TO_RADS(dir) ) );
  car_right.smult( sin( DEGS_TO_RADS(dir) ) );
  VEC patch_vel;
  patch_vel.add( car_forward, car_right );
//  patch_vel.normalize();	/* don't need this if car_forward/right */
//  are normalized already

  patch_vel.smult(-spin*TIRE_CIRCUMF); 	// rel. to car, patch is moving backwrd
  VEC patch_vel_ground;
  patch_vel_ground.add( patch_vel, car_vel );

  // now we want this vel vector to be perpendicular to surface normal
  VEC tmp;
  tmp.xprod( patch_vel_ground, surf_normal );
  v_net_u.xprod( surf_normal, tmp );
  v_net_u.normalize();

  return true;
}

bool
wheel::get_wheel_dir( VEC car_fwd, const VEC& car_up, VEC& result ) {
  VEC car_right;
  car_right.xprod( car_fwd, car_up );
  car_fwd.smult( cos(DEGS_TO_RADS(dir)) );
  car_right.smult( sin(DEGS_TO_RADS(dir)) );
  VEC Fwh;
  Fwh.add( car_fwd, car_right );

  // now it has to be perpendicular to surf_normal
  VEC tmp;
  tmp.xprod( Fwh, surf_normal );
  result.xprod( surf_normal, tmp );
  result.normalize();

  return true;
}

///////////////////////////////////////////////////////////////////
// returns the driving force vector (perpendicular to surf_normal
///////////////////////////////////////////////////////////////////
bool
wheel::get_F_wheel_vec( const VEC& car_fwd, const VEC& car_up, VEC& result ) {
  get_wheel_dir( car_fwd, car_up, result );
  result.smult( Fwheel );

  return true;
}

/*
 *
 * draws circle/disk on z=0 plan, of radius 'rad'
 * normal in +z direction
 */

void draw_circle( float rad ) {

	GLUquadricObj* quadObj;
	quadObj = gluNewQuadric();
	gluQuadricDrawStyle( quadObj, GLU_FILL );
	gluQuadricNormals ( quadObj, GLU_SMOOTH );
	gluDisk( quadObj, 0, rad, 12, 2 );
	gluDeleteQuadric( quadObj );
}


/*
 *
 * fix and improvement over the auxSolidCylinder() which does not draw
 * the cylinder centered around origin, but rather sitting on top of it
 *
 * draws a capped (at both ends) cylinder of radius 'rad' and 
 * length 'length', with its axis right on y axis, and centered on
 * origin
 */
void draw_capped_cylinder( float rad, float length ) {

	glPushMatrix();
	glScalef(1,length/2,1);
	auxSolidCylinder( rad, 2 );
	glPopMatrix();

	// upper disk
	glPushMatrix();
	glTranslatef(0,length/2,0);
	glRotatef(-90,1,0,0);
	draw_circle( rad );
	glPopMatrix();

	// lower disk
	glPushMatrix();
	glTranslatef(0,-length/2,0);
	glRotatef(90,1,0,0);
	draw_circle( rad );
	glPopMatrix();
}


////////////////////////////////////////////////////////////
// draws a wheel
////////////////////////////////////////////////////////////
void draw_wheel() {
  set_colour( 0.5, 0.5, 0.5 );
  glPushMatrix();
  draw_capped_cylinder( TIRE_RADIUS, 0.2 );
  glPopMatrix();
}

////////////////////////////////////////////////////////////
// draws the car
////////////////////////////////////////////////////////////
void draw_car() {
  Matrix R, Rt;

  glPushMatrix();
  glTranslatef( State[0], State[1], State[2] - CAR_TB/2 );
  quat_to_mat(State+3, R);
  tmtransp(Rt, R);
  glMultMatrixf((float *) Rt);

  for( int w=0; w<4; w++ ) {
    glPushMatrix();
    glTranslatef( rvec_base[w][0],rvec_base[w][1],rvec_base[w][2] );
    glRotatef( 90+wheels[w].dir, 0, 0, 1 );
    draw_wheel();
    glPopMatrix();
  }

  glScalef( CAR_SS, CAR_FB, CAR_TB );
  set_colour( 0.0, 1.0, 0.0 );
  auxSolidCube( 1.0 );
  glPopMatrix();
}


////////////////////////////////////////////////////////////////
// calculates the shock absorbers' force
////////////////////////////////////////////////////////////////
void 
wheel::calc_F_shock() {
  double len = -shock_kp*(dist - shock_l0) - shock_kd*(dist-shock_prev_l)/Dtsim;
  shock_prev_l = dist;
  shock_force = surf_normal;
  shock_force.smult(len);
}

/////////////////////////////////////////////////////////////////
// calculates the closest point info
/////////////////////////////////////////////////////////////////
bool 
wheel::get_closest_info() {
  
  track_seg* pts;
  int res;
  do {
    pts = trk.get_seg( last_seg );
    if( !pts ) {
      cout << "Error: out of track!" << endl;
      return false;
    }
    res = pts->get_closest_info( pos, closest_pnt, dist, surf_normal );
    if( res<0 )
      last_seg--;
    if( res>0 )
      last_seg++;
  } while( res );

  return true;
}


void
wheel::calc_spin( const VEC& car_vel, const VEC& car_fwd ) {
  if( mode == MODE_STATIC ) {
    spin = car_vel.size() / TIRE_RADIUS;
    if( dotprod( car_vel, car_fwd ) < 0 )
      spin *= -1;
  } else
  if( mode == MODE_SLIP ) {
    // !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! add code here
  }
}



