#include <iostream.h>

#include "lib/lib_defs.h"
#include "lib/lib_math.h"

#include "car.h"
#include "user.h"
#include "sim.h"

/////////////////// constants
const double myu_k = 0.8;
const double myu_s = 1.0;

#define TOTAL_EQNS	(72-12-12)

#define MASS		10.0		// 10 kg; ok, so it's a go-cart...


#define r_NEWTON	0
#define r_EULER		1
#define r_FNET		2
#define r_F1		3
#define r_F1COM		7
#define r_F2COM		8
#define r_F3COM		9
#define r_F4COM		10
#define r_TAU1		11
#define r_TAU2		12
#define r_TAU3		13
#define r_TAU4		14
#define r_TAUNET	15

// these are out of the eqn solver; not using kinetic friction at this
// time
#define r_F1DN		16
#define r_F1ON		20

/////////////////// globals
MAT A;
float b[MAX_EQNS];
float xx[MAX_EQNS];

FVEC vec_fwd_base = { 0, 1, 0, 0 };
FVEC vec_up_base = { 0, 0, 1, 0 };
FVEC vec_fwd;
FVEC vec_up;

bool fSLEbypass = false;
bool xprint = false;

// each wheels vector from COM to that wheel, in OCS
FVEC rvec_base[4] = {
  {-1,2,-0.5,0},
  {1,2,-0.5,0},
  {-1,-2,-0.5,0},
  {1,-2,-0.5,0}
};

// each wheel's vector from COM to that wheel, in WCS
FVEC rvec[4];

////////////////////////////////////////////////////
//	PROC:	void dyn_init_state()
//	DOES:	initializes the state vector
////////////////////////////////////////////////////
void dyn_init_state()
{
  // first block
  State[0] = 0.0;  // position
  State[1] = 5.0;
  State[2] = 1.0;

  State[3] = 1.0;  // orientation
  State[4] = 0.0;
  State[5] = 0.0;
  State[6] = 0.0;
  
  State[7] = 0.0;  // velocity
  State[8] = 0.0;
  State[9] = 0.0;

  State[10] = 0.0; // angular velocity
  State[11] = 0.0;
  State[12] = 0.0;

//  wheels[0].dir = wheels[1].dir = 90;
//  wheels[2].Fwheel = wheels[3].Fwheel = 10;
}

////////////////////////////////////////////////////
//	PROC:	void dyn_sim()
//	DOES:	takes a simulation time step
////////////////////////////////////////////////////
void dyn_sim()
{
  Matrix R, Rt, temp, I;
  static Matrix Ib = {
	 MASS/12.0*(CAR_FB*CAR_FB+CAR_TB*CAR_TB), 0, 0, 0,
	 0, MASS/12.0*(CAR_SS*CAR_SS+CAR_TB*CAR_TB), 0, 0,
	 0, 0, MASS/12.0*(CAR_SS*CAR_SS+CAR_FB*CAR_FB), 0,
	 0, 0, 0, 1};

  FVEC q_dot;
  float *x = State+0;
  float *q = State+3;
  float *x_dot = State+7;
  float *omega = State+10;

  quat_to_mat(q, R);           // compute I = R Ib R^t
  tmtransp(Rt, R);
  tmmmult(temp, Ib, Rt);
  tmmmult(I, R, temp);

  for( int i=0; i<4; i++ )
    tmmult( rvec[i], R, rvec_base[i] );
  tmmult( vec_fwd, R, vec_fwd_base );
  tmmult( vec_up, R, vec_up_base );

  // compute wheel positions
  // and fill the "closest ground info" cache!
  for( int i=0; i<4; i++ ) {
    wheels[i].pos.add( VEC(State), VEC(rvec[i]) );

    wheels[i].get_closest_info();
    wheels[i].calc_spin( VEC(State+7), vec_fwd );
    wheels[i].get_patch_vel( VEC(x_dot), VEC(vec_fwd), VEC(vec_up) );
    wheels[i].calc_F_shock();
  }

  // fill matrix A and vector b
  // clear to zero first
  for (int r=0; r<TOTAL_EQNS; r++) {
	b[r] = 0.0;
	for (int c=0; c<TOTAL_EQNS; c++) 
	  A[r][c]=0;
  }

  fill_newton(MASS);          // fill matrix
  fill_euler(omega,I);
  fill_Fnet(MASS);
  for( int r=0; r<4; r++ )
    switch( wheels[r].mode ) {
      case MODE_STATIC:
	fill_wheel_static( r, MASS );
	break;
      case MODE_SLIP:
	fill_wheel_kinetic( r );
	break;
      case MODE_AIR:
	fill_wheel_airborne( r );
	break;
    }
  fill_F_COMs();
  fill_torques();
  fill_F_downs(MASS);
  fill_F_ons();




  if( wheels[0].mode == MODE_STATIC ) {
    VEC sn = wheels[0].surf_normal;
    VEC tmp;
    VEC grav( 0, 0, -MASS*9.8 );
    tmp.project_on( grav, sn );
    VEC sd;
    sd.sub( grav, tmp );
    VEC res=sd;
    res.smult(0.25);
//    for( int i=0; i<3; i++ )
//      b[r_FNET*3+i] += sd.vec3()[i];
    for( int w=0; w<4; w++ ) {
      for( int r=0; r<3; r++ )
	b[r_F1*3+w*3+r] += res.vec3()[r];
    }
  }




  if( !fSLEbypass ) {
    for( int w=0; w<4; w++ ) {
      cout << "wheel " << w << " in ";
      switch( wheels[w].mode ) {
	case MODE_AIR:
	  cout << "AIR mode" << endl;
	  break;
	case MODE_STATIC:
	  cout << "STATIC mode" << endl;
	  break;
	case MODE_SLIP:
	  cout << "SLIP mode" << endl;
	  break;
      }
    }
    cout << endl;

    SLE_print_zoomedout();

    SLE_query();
  }


#ifdef PRINT_MATRICES
  prmat(A,TOTAL_EQNS);
  printf("b=");
  prnvec(b,TOTAL_EQNS);
#endif

  lin_eq_solv(A, b, xx, TOTAL_EQNS);     // solve for unknowns

  if( xprint ) {
    cout << "a  " << xx[0] << "," << xx[1] << "," << xx[2] << "\t";
    cout << "w.  " << xx[3] << "," << xx[4] << "," << xx[5] << endl;
    cout << "Fnet  " << xx[6] << "," << xx[7] << "," << xx[8] << endl;
    cout << "F1  " << xx[9] << "," << xx[10] << "," << xx[11] << "\t";
    cout << "F2  " << xx[12] << "," << xx[13] << "," << xx[14] << endl;
    cout << "F3  " << xx[15] << "," << xx[16] << "," << xx[17] << "\t";
    cout << "F4  " << xx[18] << "," << xx[19] << "," << xx[20] << endl;
    cout << "F1com  " << xx[21+0] << "," << xx[21+1] << "," << xx[21+2] << "\t";
    cout << "F2com  " << xx[21+3] << "," << xx[21+4] << "," << xx[21+5] << endl;
    cout << "F3com  " << xx[27+0] << "," << xx[27+1] << "," << xx[27+2] << "\t";
    cout << "F4com  " << xx[27+3] << "," << xx[27+4] << "," << xx[27+5] << endl;
    cout << "tau1  " << xx[33+0] << "," << xx[33+1] << "," << xx[33+2] << "\t";
    cout << "tau2  " << xx[33+3] << "," << xx[33+4] << "," << xx[33+5] << endl;
    cout << "tau3  " << xx[39+0] << "," << xx[39+1] << "," << xx[39+2] << "\t";
    cout << "tau4  " << xx[39+3] << "," << xx[39+4] << "," << xx[39+5] << endl;
    cout << "tauNet  " << xx[45] << "," << xx[46] << "," << xx[47] << endl;
    cout << "F1dn  " << xx[48+0] << "," << xx[48+1] << "," << xx[48+2] << "\t";
    cout << "F2dn  " << xx[48+3] << "," << xx[48+4] << "," << xx[48+5] << endl;
    cout << "F3dn  " << xx[54+0] << "," << xx[54+1] << "," << xx[54+2] << "\t";
    cout << "F4dn  " << xx[54+3] << "," << xx[54+4] << "," << xx[54+5] << endl;
    cout << "F1on  " << xx[60+0] << "," << xx[60+1] << "," << xx[60+2] << "\t";
    cout << "F2on  " << xx[60+3] << "," << xx[60+4] << "," << xx[60+5] << endl;
    cout << "F3on  " << xx[66+0] << "," << xx[66+1] << "," << xx[66+2] << "\t";
    cout << "F4on  " << xx[66+3] << "," << xx[66+4] << "," << xx[66+5] << endl;

    xprint = false;
  }

  float *x_dot_dot = xx;        // get state derivative ready
  float *omega_dot = xx+3;      
  solve_qdot(q_dot, q, omega); 
  
  // integrate the position, x
  for (int d=0; d<3; d++)
	x[d] += x_dot[d]*Dtsim + 0.5*x_dot_dot[d]*Dtsim*Dtsim;

  // integrate the orientation, q
  for (int d=0; d<4; d++)
	q[d] += q_dot[d]*Dtsim;

  qnorm(q);               // renormalize

  // integrate the velocity, x_dot
  for (int d=0; d<3; d++)
	x_dot[d] += x_dot_dot[d]*Dtsim;
  
  // integrate the angular velocity, omega
  for (int d=0; d<3; d++)
	omega[d] += omega_dot[d]*Dtsim;


  // NOW, figure out which mode each wheel is in AFTER this time step
  for( int w=0; w<4; w++ ) {
    if( wheels[w].dist <= TIRE_RADIUS )
      // check whether STATIC or SLIP mode !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
      wheels[w].mode = MODE_STATIC;
    else 
      wheels[w].mode = MODE_AIR;
  }
}

////////////////////////////////////////////////////
//	PROC:	void fill_newton(mass)
//	DOES:	Newton's equation
////////////////////////////////////////////////////
void fill_newton(float m)
{
  for (int r=0; r<3; r++) {
	A[r][r] = m;
	A[r][r_FNET*3+r] = -1;
  }
}

////////////////////////////////////////////////////
//	PROC:	void fill_euler(omega, I)
//	DOES:	Euler's equation
////////////////////////////////////////////////////
void fill_euler(float *omega, Matrix I) {
  FVEC temp1, temp2;

  // compute -omega x (I omega), store in b
  tmmult(temp1, I, omega);
  xprod(temp2, omega, temp1);
  b[r_EULER*3+0] = -temp2[0];
  b[r_EULER*3+1] = -temp2[1];
  b[r_EULER*3+2] = -temp2[2];

  // place I in A
  for (int r=0; r<3; r++)
    for (int c=0; c<3; c++)
      A[r_EULER*3+r][r_EULER*3+c] = I[r][c];

  // take tau_{net} into account  (net torque)
  for( int i=0; i<3; i++ )
    A[r_EULER*3+i][r_TAUNET*3+i] =-1;
}

////////////////////////////////////////////////////
//	PROC:	void fill_Fnet( mass )
//	DOES:	Fnet = F1com + F2com + F3com + F4com + mg 
////////////////////////////////////////////////////
void fill_Fnet( double mass ) {

  /*
  VEC tmp(0,1,0), sd, grav(0,0,-9.8*mass);
  sd.xprod( tmp, wheels[0].surf_normal );
  sd.normalize();
  tmp.project_on( grav, sd );
  tmp.normalize();
  tmp.smult(-1);
  */

  const float GRAVITY = -9.8;
  b[r_FNET*3+2] = -mass*GRAVITY;
  for( int r=0; r<3; r++) {
    A[r_FNET*3+r][r_FNET*3+r] = -1;
    A[r_FNET*3+r][r_F1COM*3+r] = 1;
    A[r_FNET*3+r][r_F2COM*3+r] = 1;
    A[r_FNET*3+r][r_F3COM*3+r] = 1;
    A[r_FNET*3+r][r_F4COM*3+r] = 1;
//    b[r_FNET*3+r] += tmp.vec3()[r];
  }
}


////////////////////////////////////////////////////
//	PROC:	void fill_wheel_kinetic()
//	DOES:	eq'n for wheel if slipping
////////////////////////////////////////////////////
void fill_wheel_kinetic( int w ) {

  VEC v_net_u = wheels[w].v_net_u;

  for( int r=0; r<3; r++ ) {
    A[r_F1*3+w*3+r][r_F1*3+w*3+r] = -1;
    b[r_F1*3+w*3+r] = -wheels[w].shock_force.vec3()[r];
    for( int c=0; c<3; c++ )
      A[r_F1*3+w*3+r][r_TAUNET*3+w*3+c] = 
	      -myu_k * v_net_u.vec3()[r] * v_net_u.vec3()[c];
  }
}
    

////////////////////////////////////////////////////
//	PROC:	void fill_wheel_static()
//	DOES:	eq'n for wheel if NOT slipping
////////////////////////////////////////////////////
void fill_wheel_static( int w, double mass ) {

  VEC sd, tmp, fwd;
  wheels[w].get_wheel_dir( vec_fwd, vec_up, fwd );
  sd.xprod( wheels[w].surf_normal, fwd );


  /*
  tmp.project_on( grav, wheels[w].surf_normal );
  VEC Fsd;
  Fsd.sub( grav, tmp );
  tmp = rvec[w];
  tmp.normalize();
  VEC Fonwheel;
  Fonwheel.project_on( Fsd, tmp );
  VEC Fonwhsd;
  Fonwhsd.project_on( Fonwheel, sd );
    Fonwhsd.project_on( grav, sd );
  VEC res;
  res.sub( Fonwhsd, wheels[w].shock_force );
  */


  tmp = VEC(rvec[w]);
  tmp.normalize();
  tmp.smult(-1);
  VEC tmp2,tmp3;
  tmp3.project_on(tmp, VEC(vec_up) );
  tmp2.sub( tmp, tmp3 );
  tmp2.normalize();
  double cos_theta = dotprod( tmp2, tmp );
  double theta = acos( cos_theta );
  double x = tan(theta)*tan(theta)/1;
  double multiplier = x+1;

  VEC grav( 0, 0, mass * -9.8 );
  grav.smult(0.25);
  tmp2.project_on( sd, wheels[w].surf_normal );
  tmp3.sub( sd, tmp2 );
  tmp3.normalize();
  tmp.project_on( grav, tmp3 );
  tmp.smult(multiplier);

  VEC Fwheel;
  wheels[w].get_F_wheel_vec( VEC(vec_fwd), VEC(vec_up), Fwheel );
  tmp2.sub( tmp, Fwheel );

  VEC res;
  res.sub( tmp2, wheels[w].shock_force );

  /*
  grav.smult( 0.25 );
  VEC Fwheel;
  wheels[w].get_F_wheel_vec( VEC(vec_fwd), VEC(vec_up), Fwheel );
  tmp.add( grav, Fwheel );

  double len = dotprod( tmp, sd );
  tmp = sd;
  tmp.smult( len );
  VEC res;
  res.sub( tmp, wheels[w].shock_force );
  */

  /*
  VEC res;
  res  = wheels[w].shock_force;
  res.smult(-1);
  */

  for( int r=0; r<3; r++ ) {
    for( int c=0; c<3; c++ ) {
      A[r_F1*3+w*3+r][r_F1*3+w*3+r] = -1;
//      A[r_F1*3+w*3+r][r_F1COM*3+w*3+c] = sd.vec3()[r]*sd.vec3()[c];
      b[r_F1*3+w*3+r] = res.vec3()[r];
    }
  }
}

////////////////////////////////////////////////////
//	PROC:	void fill_wheel_airborne()
//	DOES:	eq'n for wheel if airborne
////////////////////////////////////////////////////
void fill_wheel_airborne( int w ) {
  for( int r=0; r<3; r++ ) {
    A[r_F1*3+w*3+r][r_F1*3+w*3+r] = -1;
  }
}

////////////////////////////////////////////////////
//	PROC:	void fill_F_COMs()
//	DOES:	eq'n for force exerted by wheels on COM
////////////////////////////////////////////////////
void fill_F_COMs() {
  for( int w=0; w<4; w++ ) {
    VEC ru = rvec[w];		// COM to wheel
    ru.smult(-1);		// wheel to COM
    ru.normalize();

    for( int r=0; r<3; r++ ) {
      for( int c=0; c<3; c++ ) {
	A[r_F1COM*3+w*3+r][r_F1*3+w*3+c] = ru.vec3()[r]*ru.vec3()[c];
      }
    }

    for( int i=0; i<3; i++ )
      A[r_F1COM*3+w*3+i][r_F1COM*3+w*3+i] = -1;
  }
}

////////////////////////////////////////////////////
//	PROC:	void fill_torques()
//	DOES:	eq'n for torques exerted by wheels w.r.t. COM
////////////////////////////////////////////////////
void fill_torques() {
  for( int w=0; w<4; w++ ) {

    Matrix rmat;
    xprodmat( rvec[w], rmat );

    for( int r=0; r<3; r++ ) {
      for( int c=0; c<3; c++ ) {
	A[r_TAU1*3+w*3+r][r_F1COM*3+w*3+c] = -rmat[r][c];
	A[r_TAU1*3+w*3+r][r_F1*3+w*3+c] = rmat[r][c];
      }
    }

    for( int i=0; i<3; i++ )
      A[r_TAU1*3+w*3+i][r_TAU1*3+w*3+i] = -1;
  }

  // now do net torque
  for( int i=0; i<3; i++ ) {
    A[r_TAUNET*3+i][r_TAU1*3+i] = 1;
    A[r_TAUNET*3+i][r_TAU2*3+i] = 1;
    A[r_TAUNET*3+i][r_TAU3*3+i] = 1;
    A[r_TAUNET*3+i][r_TAU4*3+i] = 1;
    A[r_TAUNET*3+i][r_TAUNET*3+i] = -1;
  }
}


/////////////////////////////////////////////////////////////////////////
//	PROC:	void fill_F_downs(MASS)
//	DOES:	eq'n for force exerted by each wheel on ground
/////////////////////////////////////////////////////////////////////////
void fill_F_downs( double mass ) {
  VEC grav( 0, 0, mass * -9.8 );

  for( int w=0; w<4; w++ ) {
    double len = 0.25 * dotprod( wheels[w].surf_normal, grav );
    VEC q_grav_normal;
    q_grav_normal = wheels[w].surf_normal;
    q_grav_normal.smult( len );

    for( int i=0; i<3; i++ )
      b[r_F1DN*3+w*3+i] = q_grav_normal.vec3()[i];

    VEC beta;
    beta.xprod( wheels[w].surf_normal, rvec[w] );

    for( int r=0; r<3; r++ ) {
      for( int c=0; c<3; c++ ) {
	A[r_F1DN*3+w*3+r][r_TAUNET*3+c] = 
	  		beta.vec3()[c]*wheels[w].surf_normal.vec3()[r];
      }
    }

    for( int i=0; i<3; i++ )
      A[r_F1DN*3+w*3+i][r_F1DN*3+w*3+i] = -1;
  }
}

/////////////////////////////////////////////////////////////////////////
//	PROC:	void fill_F_ons()
//	DOES:	eq'n for force exerted on a wheel by the other 3 wheels
/////////////////////////////////////////////////////////////////////////
void fill_F_ons() {
  for( int w=0; w<4; w++ ) {

    for( int C=0; C<4; C++ )
      if( w!=C ) {

	VEC ru;
	ru.sub( wheels[w].pos, wheels[C].pos );
	ru.normalize();

	for( int r=0; r<3; r++ ) {
	  for( int c=0; c<3; c++ ) {
	    A[r_F1ON*3+w*3+r][r_F1*3+C*3+c] = ru.vec3()[r]*ru.vec3()[c];
	  }
	}
      }

    for( int i=0; i<3; i++ )
      A[r_F1ON*3+w*3+i][r_F1ON*3+w*3+i] = -1;
  }
}



// R,C are the coords of the sector (stride 3)
bool is_Asector_nonzero( int R, int C ) {
  R*=3;
  C*=3;

  for( int r=0; r<3; r++ )
    for( int c=0; c<3; c++ )
      if( A[R+r][C+c] )
	return true;

  return false;
}

// R is the coord of the sector (stride 3)
bool is_Bsector_nonzero( int R ) {
  R*=3;

  for( int r=0; r<3; r++ )
    if( b[R+r] )
	return true;

  return false;
}


void SLE_print_zoomedout() {

  cout << endl << endl;
  cout << "    0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0"
    << endl;

  int no_sectors = TOTAL_EQNS/3;

  for( int R=0; R<no_sectors; R++ ) {
    if( is_Bsector_nonzero( R ) )
      cout << "b ";
    else
      cout << "  ";

    cout << "| ";

    for( int C=0; C<no_sectors; C++ ) {
      if( is_Asector_nonzero( R, C ) )
	cout << "x ";
      else
	cout << "  ";
    }
    cout << endl;
  }
}

void SLE_print_Asector( int R, int C ) {
  R*=3;
  C*=3;

  for( int r=0; r<3; r++ ) {
    for( int c=0; c<3; c++ )
      cout << A[R+r][C+c] << "\t";
    cout << endl;
  }
}

void SLE_print_Bsector( int R ) {
  R*=3;

  for( int r=0; r<3; r++ )
    cout << b[R+r] << endl;
}


void SLE_query() {

  for(;;) {

    cout << endl;
    cout << "Enter SLE query: " << endl;
    char cmd;
    int R, C;
    cin >> cmd;
    switch( cmd ) {
      case 'Q':
	fSLEbypass = true;
	return;
      case 'q':
	return;
      case 'a':
      case 'A':
	cin >> R >> C;
	SLE_print_Asector( R, C );
	break;
      case 'b':
      case 'B':
	cin >> R;
	SLE_print_Bsector( R );
	break;
      case 'x':
	xprint = true;
	return;
      default:
	cout << endl;
	cout << "invalid command; use 'Q', 'q', 'a', or 'b'." << endl;
	break;
    }
  }
}


