#include <iostream.h>
#include <fstream.h>
#include <math.h>

#include <GL/gl.h>
#include <GL/glu.h>

#include "glaux.h"

#include "vec_math.h"
#include "track.h"
#include "user.h"

const double TRACK_WIDTH=5;

/*
 * track_seg::get_tilt( double len )
 *
 * returns the tilt at a distance of 'len' from the beginning of the
 * track segment
 *
 * tilt>0 == tilts right 'tilt' degrees, when looking down direction of travel
 * tilt<0 == tilts left 'tilt' degrees, when looking down direction of travel
 */
inline
double
track_seg::get_tilt( double len ) {
  // clamp the values
  if( len < 0 )
    len = 0;
  if( len > length )
    len = length;

  return tilt_init + (tilt_fin - tilt_init)*(len/length);
}

/*
 * track_seg::get_untilted_normal( double len, VEC& n, VEC& t )
 *
 * returns the normal and the tangent of the track segment, at 
 * distance 'len' from the beginning of segment, 
 * ignoring the 'tilt' at that point 
 * (won't necessarily be the 0,0,1 vector since even a straight segment
 * can be on a downward slope)
 *
 * the returned normal, tangent have been normalized
 */
void
track_seg::get_untilted_normal_tangent( double len, VEC& norm, VEC& tngnt ) {
  if( radius ) {
    // consider curved segments here
    VEC tmp, i, j;
    if( len ) {
      double sweep = len/length * get_sweep();
	j = get_orient_norm();
	i = dir_init;
	j.smult( radius * sin( DEGS_TO_RADS(sweep) ) );
	i.smult( radius * cos( DEGS_TO_RADS(sweep) ) );
	tngnt.add( i, j );
	tmp.xprod( tngnt, dir_init );
	if( sweep > 180 && sweep < 360 )
	  tmp.smult(-1);
      if( tmp.size() < 0.001 ) {
	// the tangents are close to parallel
	VEC tmp2;
	j = get_orient_norm();
	i = dir_init;
	j.smult( radius * sin( DEGS_TO_RADS(90) ) );
	i.smult( radius * cos( DEGS_TO_RADS(90) ) );
	tmp2.add( i, j );
	tmp.xprod( tmp2, dir_init );
      }
      norm.xprod( tngnt, tmp );
      tngnt.normalize();		// ???
    } else {
      tngnt = dir_init;
      norm = get_orient_norm();
    }
  } else {
    // consider straight segments here

    norm = get_orient_norm();
    tngnt = dir_init;
  }
}




/*
 * track_seg::get_normal_tangent( double len, VEC& n, VEC& t )
 *
 * returns the normal and the tangent of the track segment at 
 * distance 'len' from the beginning of segment
 *
 * the normal and tangent returned is already normalized
 *
 * return value: true if successful
 */
bool
track_seg::get_normal_tangent( double len, VEC& n, VEC& t ) {
  // clamp values
  if( len < 0 )
    len = 0;
  if( len > length )
    len = length;

  double tilt = get_tilt( len );

  VEC unorm, i, j;
  get_untilted_normal_tangent( len, unorm, t );
  i.xprod( t, unorm );
  i.normalize();
  j = unorm;
  j.normalize();
  i.smult( sin( DEGS_TO_RADS(tilt) ) );
  j.smult( cos( DEGS_TO_RADS(tilt) ) );
  n.add( i, j );
  return true;
}

/*
 * track_seg::get_sweep()
 *
 * if the segment is curved, returns the number of degrees that the
 * curve sweeps out; else returns 0 for straights
 */
double
track_seg::get_sweep() {
  if( !radius )
    return 0;

  double radians = length/radius;
  return RADS_TO_DEGS(radians);	// we want the answer in degrees
}


/*
 * track_seg::draw()
 *
 * draws itself
 */
#define CURVE_FRAGS_pul	(0.5)	// pul == per unit length
#define DRAW_NORMALS
void
track_seg::draw() {

#ifdef DRAW_NORMALS
  set_colour( 0.3, 0.3, 0.3 );
#endif

  if( !radius && tilt_init == tilt_fin ) {
    // this is a straight segment
    VEC a,b,c,d;
    VEC e,f,g,h;
    VEC n,t;
    get_pos_extent( 0, a, b );
    get_pos_extent( length, c, d );
    e.add(a,b);
    f.sub(a,b);
    g.add(c,d);
    h.sub(c,d);

    glBegin(GL_QUAD_STRIP);
      get_normal_tangent( 0, n, t );
      glNormal3fv(n.vec3());
      glVertex3fv(e.vec3());
      glNormal3fv(n.vec3());
      glVertex3fv(f.vec3());
      glNormal3fv(n.vec3());
      glVertex3fv(g.vec3());
      glNormal3fv(n.vec3());
      glVertex3fv(h.vec3());
    glEnd();
  }
  else {
    // this is a curved segment
    int total_frags = (int)(ceil(length * CURVE_FRAGS_pul));
    double delta = length/(double)total_frags;
    double len = 0;
    glBegin(GL_QUAD_STRIP);
    for( int frag=0; frag<=total_frags; frag++, len+=delta ) {
      VEC a,b,c,d;
      VEC n,t;
      get_pos_extent( len, a, b );
      get_normal_tangent( len, n, t );
      c.add(a,b);
      d.sub(a,b);
      if( c.isZero() || d.isZero() ) {
	cout << "NoW!!!" << endl;
      }
	glNormal3fv( n.vec3() );
	glVertex3fv( c.vec3() );
	glNormal3fv( n.vec3() );
	glVertex3fv( d.vec3() );
    }
    glEnd();

#ifdef DRAW_NORMALS
    set_colour( 1.0, 0.0, 0.0 );
    len = 0;
    for( int frag=0; frag<=total_frags; frag++, len+=delta ) {
      VEC a,b,c,d;
      VEC n,t;
      get_pos( len, a );
      get_normal_tangent( len, n, t );
//      n.smult(0.1);
      c.add(a,n);
      glBegin(GL_LINES);
	glNormal3f( 0, 0, 1 );
	glVertex3fv( a.vec3() );
	glNormal3f( 0, 0, 1 );
	glVertex3fv( c.vec3() );
      glEnd();
    }
#endif

  }
}


/*
 * track_seg::get_pos( double len, VEC& p )
 *
 * gives the 3d position of a point on a segment a distance of 'len'
 * away from the segment start
 *
 * the result is placed in 'p'
 */
bool 
track_seg::get_pos( double len, VEC& p ) {
  // clamp values
  if( len < 0 )
    len = 0;
  if( len > length )
    len = length;

  if( radius ) {
    // do curved segment
    double sweep = len/length * get_sweep();

    VEC i, j, tmp;
    i = dir_init;
    j = get_orient_norm();
    j.smult( radius * (-cos( DEGS_TO_RADS(sweep) )+1) );
    i.smult( radius * sin( DEGS_TO_RADS(sweep) ) );
    tmp.add( i, j );
    p.add( tmp, pos );
  } else {
    // do straight segment
    VEC tmp=dir_init;
    tmp.smult( len );
    p.add( pos, tmp );
  }
  return true;
}


/*
 * track_seg::get_pos_extent( double len, VEC& p, VEC& right )
 *
 * gives the 3d position of a point on a segment a distance of 'len'
 * away from the segment start, as well as a vector, 'right', which when
 * added to 'p', gives the 3d position of the right edge of the track
 *
 * the result is placed in 'p'
 */
bool
track_seg::get_pos_extent( double len, VEC& p, VEC& right ) {
  // clamp values
  if( len < 0 )
    len = 0;
  if( len > length )
    len = length;

  VEC n, t;
  get_pos( len, p );
  get_normal_tangent( len, n, t );
  right.xprod( t, n );
  right.normalize();
  right.smult( TRACK_WIDTH/2 );
  return true;
}


/*
 *
 *
 * result already normalized
 */
VEC 
track_seg::get_orient_norm() {
  VEC tmp, i, j;

  tmp.xprod( dir_init, VEC(0,0,1) );
  j.xprod( tmp, dir_init );
  j.normalize();
  i.xprod( dir_init, j );
  i.smult( sin( DEGS_TO_RADS(orient) ) );
  j.smult( cos( DEGS_TO_RADS(orient) ) );
  tmp.add( i, j );
  return tmp;
}


////////////////////////////////////////////////////////////
bool
track::load( const char* szFile ) {
  fstream fin( szFile, ios::in );
  if( !fin )
    return false;

  fin >> n;
  segs = new (track_seg*)[n];
  if( !segs )
    return false;

  VEC p(0,0,0), tngnt(0,1,0);

  for( int i=0; i<n; i++ ) {
    double length, rad, t1, t2, ornt;

    fin >> length >> rad >> ornt >> t1 >> t2;

    segs[i] = new track_seg( length, rad, t1, t2, ornt, p, tngnt );
    if( !segs[i] )
      return false;

    VEC tmp;
    segs[i]->get_pos( length, p );
    segs[i]->get_normal_tangent( length, tmp, tngnt );
  }
}


// returns -1 if  sweep < 0
// returns 0  if  sweep in range
// returns 1  if  sweep > 1
int
track_seg::get_closest_info( const VEC& p1, 
    				VEC& p2, 
				double& dist, 
				VEC& normal 
				) {
  // "Mucho expensivo!"
  // could be worse: was gonna do an iterative solution... <Yikes!>

  if( radius ) {	// curved segment!
    VEC n, t;
    get_untilted_normal_tangent( 0, n, t );	// already normalized
    VEC s;
    s.xprod( t, n );	// perp. vector RIGHT of curve
    VEC tmp;
    tmp.sub( pos, p1 );	// vector from tire.pos to seg start pos
    double len = dotprod( tmp, s );
    VEC to_plane = s;
    to_plane.smult( len );	// this vector now moves us from p1 to p1'
    				// on curve plane surface
    VEC p1p;
    p1p.add( p1, to_plane );

    tmp = n;
    tmp.smult( radius );
    VEC center;
    center.add( tmp, pos );	// center of the curve

    VEC p1pp;
    tmp.sub( p1p, center );
    tmp.normalize();
    tmp.smult( radius );
    p1pp.add( tmp, center );

    tmp.sub( pos, center );
    VEC tmp2;
    tmp2.sub( p1pp, center );
    len = dotprod( tmp, tmp2 )/radius/radius;

    double sweep = acos(len) / (length/radius);
    if( sweep<0 || sweep>1 ) {
      if( sweep< 0 )
	return -1;
      else
	return 1;
    }

    VEC tn, tt;
    get_normal_tangent( sweep*length, tn, tt );
    VEC r;
    r.xprod( tt, tn );		// already normalized
    tmp.sub( p1pp, p1 );
    len = dotprod( tmp, r );
    tmp = r;
    tmp.smult( -len );

    p2.add( p1pp, tmp );
    tmp.sub( p2, p1 );
    dist = tmp.size();
    tmp.sub( p1, p1pp );
    if( dotprod( tn, tmp ) < 0 )
      dist*=-1;
    normal = tn;
  } else {
    VEC n, t;
    get_untilted_normal_tangent( 0, n, t );	// already normalized
    VEC s;
    s.xprod( t, n );	// perp. vector 
    VEC tmp;
    tmp.sub( pos, p1 );	// vector from tire.pos to seg start pos
    double len = dotprod( tmp, n );
    VEC to_plane = n;
    to_plane.smult( len );	// this vector now moves us from p1 to p1'
    				// on plane surface
    VEC p1p;
    p1p.add( p1, to_plane );
    tmp.sub( pos, p1p );
    len = dotprod( tmp, s );
    VEC to_line = s;
    to_line.smult( len );
    VEC p1pp;
    p1pp.add( p1p, to_line );
    tmp.sub( p1pp, pos );	// was p1p?
    double sweep = tmp.size()/length;

    if( sweep<0 || sweep>1 ) {
      if( sweep<0 )
	return -1;
      else
	return 1;
    }

    VEC tn, tt;
    get_normal_tangent( sweep*length, tn, tt );
    VEC r;
    r.xprod( tt, tn );		// already normalized
    tmp.sub( p1, p1pp );
    len = dotprod( tmp, r );
    tmp = r;
    tmp.smult( len );

    p2.add( p1pp, tmp );
    tmp.sub( p2, p1 );
    dist = tmp.size();
    tmp.sub( p1, p1pp );
    if( dotprod( tn, tmp ) < 0 )
      dist*=-1;
    normal = tn;
  }

  return 0;
}

